/* Copyright (C) 2011, 2012, 2013 by Bundesanstalt für Gewässerkunde
 * Software engineering by Intevation GmbH
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */

package org.dive4elements.river.exports.minfo;

import org.apache.log4j.Logger;
import org.jfree.data.xy.XYSeries;
import org.w3c.dom.Document;

import org.dive4elements.artifactdatabase.state.ArtifactAndFacet;
import org.dive4elements.artifactdatabase.state.Facet;
import org.dive4elements.river.artifacts.D4EArtifact;
import org.dive4elements.river.artifacts.model.FacetTypes;
import org.dive4elements.river.artifacts.model.WKms;
import org.dive4elements.river.artifacts.model.minfo.BedDiffYearResult;
import org.dive4elements.river.artifacts.model.minfo.BedDifferencesResult;
import org.dive4elements.river.artifacts.model.minfo.MorphologicWidth;
import org.dive4elements.river.exports.StyledSeriesBuilder;
import org.dive4elements.river.exports.fixings.FixChartGenerator;
import org.dive4elements.river.exports.process.KMIndexProcessor;
import org.dive4elements.river.exports.process.Processor;
import org.dive4elements.river.exports.process.WOutProcessor;
import org.dive4elements.river.jfree.Bounds;
import org.dive4elements.river.jfree.DoubleBounds;
import org.dive4elements.river.jfree.RiverAnnotation;
import org.dive4elements.river.jfree.StyledXYSeries;
import org.dive4elements.river.utils.DataUtil;
import org.dive4elements.river.utils.RiverUtils;


public class BedDifferenceYearGenerator
extends FixChartGenerator
implements FacetTypes
{
    public enum YAXIS {
        D(0), M(1), H(2),  dW(3), W(4);

        protected int idx;

        private YAXIS(int c) {
            idx = c;
        }
    }

    /** The logger that is used in this generator. */
    private static Logger logger = Logger.getLogger(BedDifferenceYearGenerator.class);

    public static final String I18N_CHART_TITLE = "chart.beddifference.year.title";
    public static final String I18N_XAXIS_LABEL = "chart.beddifference.xaxis.label";
    public static final String I18N_YAXIS_LABEL = "chart.beddifference.yaxis.label.diff";
    public static final String I18N_SECOND_YAXIS_LABEL = "chart.beddifference.yaxis.label.morph";
    public static final String I18N_THIRD_YAXIS_LABEL = "chart.beddifference.yaxis.label.heights";

    public static final String I18N_CHART_TITLE_DEFAULT = "Sohlenhöhen Differenz";
    public static final String I18N_XAXIS_LABEL_DEFAULT = "Fluss-Km";
    public static final String I18N_YAXIS_LABEL_DEFAULT = "delta S [m]";
    public static final String I18N_SECOND_YAXIS_LABEL_DEFAULT = "Breite [m]";
    public static final String I18N_THIRD_YAXIS_LABEL_DEFAULT = "Höhe [m]";
    public static final String I18N_DW_YAXIS_LABEL_DEFAULT  =
            "delta W [cm]";
    public static final String I18N_DW_YAXIS_LABEL =
            "chart.fixings.longitudinalsection.yaxis.label";
    private static final String I18N_W_YAXIS_LABEL =
        "chart.longitudinal.section.yaxis.label";
    private static final String I18N_W_YAXIS_LABEL_DEFAULT = "W [NN + m]";

    @Override
    protected YAxisWalker getYAxisWalker() {
        return new YAxisWalker() {

            @Override
            public int length() {
                return YAXIS.values().length;
            }

            @Override
            public String getId(int idx) {
                YAXIS[] yaxes = YAXIS.values();
                return yaxes[idx].toString();
            }
        };
    }

    @Override
    public void doOut(ArtifactAndFacet bundle, Document attr, boolean visible) {
        String name = bundle.getFacetName();

        logger.debug("doOut: " + name);

        if (name == null) {
            logger.error("No facet name for doOut(). No output generated!");
            return;
        }

        Facet facet = bundle.getFacet();

        if (facet == null) {
            return;
        }

        if (getXBounds(0) != null && getDomainAxisRange() != null) {
            Bounds bounds =
                calculateZoom(getXBounds(0), getDomainAxisRange());
            context.putContextValue("startkm", bounds.getLower());
            context.putContextValue("endkm", bounds.getUpper());
        }
        else if (getXBounds(0) != null && getDomainAxisRange() == null) {
            context.putContextValue("startkm", getXBounds(0).getLower());
            context.putContextValue("endkm", getXBounds(0).getUpper());
        }
        else if (getXBounds(0) == null && getDomainAxisRange() == null) {
            BedDifferencesResult data = (BedDifferencesResult)bundle.getData(context);
            context.putContextValue("startkm", data.getKms().min());
            context.putContextValue("endkm", data.getKms().max());
        }
        else if (getXBounds(0) == null && getDomainAxisRange() != null){
            BedDifferencesResult data = (BedDifferencesResult)bundle.getData(context);
            Bounds b = new DoubleBounds(data.getKms().min(), data.getKms().max());
            Bounds bounds =
                calculateZoom(b, getDomainAxisRange());
            context.putContextValue("startkm", bounds.getLower());
            context.putContextValue("endkm", bounds.getUpper());
        }
        Processor processor = new KMIndexProcessor();
        Processor woutp = new WOutProcessor();
        if (name.equals(BED_DIFFERENCE_YEAR)) {
            doBedDifferenceYearOut(
                (BedDiffYearResult) bundle.getData(context),
                bundle, attr, visible);
        }
        else if (name.equals(BED_DIFFERENCE_MORPH_WIDTH)) {
            doBedDifferenceMorphWidthOut(
                (BedDiffYearResult) bundle.getData(context),
                bundle, attr, visible);
        }
        else if (name.equals(BED_DIFFERENCE_YEAR_HEIGHT1)) {
            doBedDifferenceHeightsOut(
                (BedDiffYearResult)bundle.getData(context),
                bundle, attr, visible, 0);
        }
        else if (name.equals(BED_DIFFERENCE_YEAR_HEIGHT2)) {
            doBedDifferenceHeightsOut(
                (BedDiffYearResult)bundle.getData(context),
                bundle, attr, visible, 1);
        }
        else if (name.equals(BED_DIFFERENCE_YEAR_FILTERED)) {
            doBedDifferenceYearOut(
                (BedDiffYearResult) bundle.getData(context),
                bundle, attr, visible);
        }
        else if (name.equals(BED_DIFFERENCE_YEAR_HEIGHT1_FILTERED)) {
            doBedDifferenceHeightsOut(
                (BedDiffYearResult)bundle.getData(context),
                bundle, attr, visible, 0);
        }
        else if (name.equals(BED_DIFFERENCE_YEAR_HEIGHT2_FILTERED)) {
            doBedDifferenceHeightsOut(
                (BedDiffYearResult)bundle.getData(context),
                bundle, attr, visible, 1);
        }
        else if (name.equals(MORPHOLOGIC_WIDTH)) {
            doMorphologicWidthOut(
                (MorphologicWidth)bundle.getData(context),
                bundle,
                attr,
                visible,
                0);
        }
        else if (processor.canHandle(name)) {
            processor.doOut(this, bundle, attr, visible, YAXIS.dW.idx);
        }
        else if (name.equals(LONGITUDINAL_ANNOTATION)) {
            doAnnotations(
                (RiverAnnotation) bundle.getData(context),
                 bundle,
                 attr,
                 visible);
        }
        else if (woutp.canHandle(name)) {
            woutp.doOut(this, bundle, attr, visible, YAXIS.W.idx);
        }
        else if (name.equals(W_DIFFERENCES)) {
            doWDifferencesOut(
                (WKms) bundle.getData(context),
                bundle,
                attr,
                visible);
        }
        else {
            logger.warn("Unknown facet name " + name);
        }
    }

    private void doMorphologicWidthOut(
        MorphologicWidth data,
        ArtifactAndFacet bundle,
        Document attr,
        boolean visible,
        int i) {
        XYSeries series = new StyledXYSeries(bundle.getFacetDescription(), attr);
        StyledSeriesBuilder.addPoints(series, data.getAsArray(), true);

        addAxisSeries(series, YAXIS.M.idx, visible);
    }

    private void doBedDifferenceHeightsOut(
        BedDiffYearResult data,
        ArtifactAndFacet bundle,
        Document attr,
        boolean visible,
        int idx) {
         logger.debug("doBedDifferenceYearOut()");

        XYSeries series = new StyledXYSeries(bundle.getFacetDescription(), attr);
        if (idx == 0) {
            StyledSeriesBuilder.addPoints(series, data.getHeights1Data(), true);
        }
        else {
            StyledSeriesBuilder.addPoints(series, data.getHeights2Data(), true);
        }

        addAxisSeries(series, YAXIS.H.idx, visible);
    }

    @Override
    protected String getDefaultChartTitle() {
        return msg(I18N_CHART_TITLE, I18N_CHART_TITLE_DEFAULT);
    }

    @Override
    protected String getDefaultXAxisLabel() {
        return msg(I18N_XAXIS_LABEL,
                   I18N_XAXIS_LABEL_DEFAULT,
                   new Object[] { getRiverName() });
    }

    @Override
    protected String getDefaultYAxisLabel(int pos) {
        String label = "default";
        D4EArtifact flys = (D4EArtifact) master;
        if (pos == YAXIS.D.idx) {
            label = msg(I18N_YAXIS_LABEL, I18N_YAXIS_LABEL_DEFAULT);
        }
        else if (pos == YAXIS.M.idx) {
            label = msg(I18N_SECOND_YAXIS_LABEL, I18N_SECOND_YAXIS_LABEL_DEFAULT);
        }
        else if (pos == YAXIS.H.idx) {
            label = msg(I18N_THIRD_YAXIS_LABEL,
                I18N_THIRD_YAXIS_LABEL_DEFAULT,
                new Object[]
                    { RiverUtils.getRiver(flys).getWstUnit().getName() });
        }
        else if (pos == YAXIS.dW.idx) {
            label = msg(I18N_DW_YAXIS_LABEL, I18N_DW_YAXIS_LABEL_DEFAULT);
        }
        else if (pos == YAXIS.W.idx) {
            return msg(I18N_W_YAXIS_LABEL, I18N_W_YAXIS_LABEL_DEFAULT,
                new Object[]
                    { RiverUtils.getRiver(flys).getWstUnit().getName() });
        }

        return label;
    }

    protected void doBedDifferenceYearOut(BedDiffYearResult data,
        ArtifactAndFacet aandf, Document theme, boolean visible) {
        logger.debug("doBedDifferenceYearOut()");

        XYSeries series = new StyledXYSeries(aandf.getFacetDescription(), theme);
        StyledSeriesBuilder.addPoints(series, data.getDifferencesData(), true);

        addAxisSeries(series, YAXIS.D.idx, visible);
    }

    protected void doBedDifferenceMorphWidthOut(BedDiffYearResult data,
        ArtifactAndFacet aandf, Document theme, boolean visible) {
        logger.debug("doBedDifferencesMorphWidthOut()");

        XYSeries series = new StyledXYSeries(aandf.getFacetDescription(), theme);
        StyledSeriesBuilder.addPoints(series, data.getMorphWidthData(), true);

        addAxisSeries(series, YAXIS.M.idx, visible);
    }

    protected void doWDifferencesOut(
        WKms       wkms,
        ArtifactAndFacet aandf,
        Document   theme,
        boolean    visible
    ) {
        if (wkms == null) {
            logger.warn("No data to add to WDifferencesChart.");
            return;
         }

        XYSeries series = new StyledXYSeries(aandf.getFacetDescription(), theme);
        StyledSeriesBuilder.addPoints(series, wkms);

        addAxisSeries(series, YAXIS.D.idx, visible);
        if (DataUtil.guessWaterIncreasing(wkms.allWs())) {
            setInverted(true);
        }
    }
}
